import hashlib
import json
from pathlib import Path

from pycopy.logging import log


def hash_file(path: Path) -> str:
    with open(path, 'rb') as file:
        hash_value = hashlib.file_digest(file, "md5")

    return hash_value.hexdigest()


class HashTracker:
    def __init__(self, relative_to: Path):
        self.relative_to = relative_to
        self.hashes = {}

    def get_hash(self, path: Path) -> str:
        try:
            return self.hashes[path]
        except KeyError:
            return None

    def update_hash(self, path: Path, other_tracker):
        self._set_hash(path, other_tracker.get_hash(path))

    def _set_hash(self, path: Path, value: str):
        self.hashes[path] = value

    def _get_combined_hash(self, paths: list) -> str:
        hash_value = hashlib.md5(usedforsecurity=False)
        for path in paths:
            new_value = self.get_hash(path.relative_to(self.relative_to))
            if new_value is None:
                raise RuntimeError("Hash could not be calculated (I apparently forgot to check some file)")
            hash_value.update(new_value.encode("ascii"))
        return hash_value.hexdigest()

    def _scan_file(self, path: Path):
        relative_path = path.relative_to(self.relative_to)

        if path.is_dir():
            subpaths = list(path.iterdir())
            for subpath in subpaths:
                self._scan_file(subpath)
            self._set_hash(relative_path, self._get_combined_hash(subpaths))
        else:
            self._set_hash(relative_path, hash_file(path))

    @classmethod
    def from_file(cls, path: Path):
        log(f"Calculating hashes for {path}")
        tracker = HashTracker(path)
        tracker._scan_file(path)
        return tracker

    @classmethod
    def from_serialized(cls, relative_to: Path, string: str):
        hashes = json.loads(string)
        tracker = HashTracker(relative_to)

        for path, hash_value in hashes.items():
            tracker.hashes[Path(path)] = hash_value

        return tracker

    def serialise(self) -> str:
        hashes = {}

        for path, hash_value in self.hashes.items():
            hashes[str(path)] = hash_value

        return json.dumps(hashes, indent=4)



import os

_right_padding = 20

def _trim_to_terminal_width(string: str):
    string = str(string)
    try:
        terminal_width = os.get_terminal_size().columns - _right_padding
    except OSError:
        return string

    if len(string) <= terminal_width:
        return string

    return string[:terminal_width // 2 - 3] + "..." + string[len(string) - terminal_width//2:]


def add_color(color, string):
    """Adds color to text on the console (using ANSI escape codes)"""
    return f"\x1b[38;5;{color}m{string}\x1b[m"


def print_temp(string):
    """Print a temporary string to the console"""
    hide_temp()
    print(_trim_to_terminal_width(string), end="", flush=True)


def hide_temp():
    """Remove the current temporary string from the console"""
    print("\x1b[1G\x1b[2K", end="", flush=True)



class Color:
    """
    Allows for tracking color properties of text stored as an array.
    The code property stores the ANSI escape codes color code for the color.
    code=None indicates that the color should be reset
    """

    def __init__(self, code):
        self.code = code

    def ansi_escape_code(self):
        """Gets the ANSI escape code sequence for this color object"""
        if self.code is None:
            return "\x1b[m"
        else:
            return f"\x1b[38;5;{self.code}m"


def log(*message, use_color=True):
    """
    Log the specified messages to the console output.
    Resolves colors specified as Color objects.
    """

    def process_message_chunk(chunk):
        if isinstance(chunk, Color):
            if not use_color: return ""
            return chunk.ansi_escape_code()
        return chunk

    message = ("[", Color(2), "pycopy", Color(None), "] ", *message, Color(None))

    print(*(process_message_chunk(chunk) for chunk in message), sep="")
